from typing import Callable

import webviz_core_components as wcc
from dash import html


def options_layout(get_uuid: Callable) -> wcc.Selectors:
    """The options part of the menu"""
    options_uuid = get_uuid("options")
    return wcc.Selectors(
        id=get_uuid("options_layout"),
        label="Options",
        children=[
            html.Div(
                id={"id": options_uuid, "element": "statistical_options"},
                children=[
                    wcc.RadioItems(
                        id={"id": options_uuid, "element": "statistical_option"},
                        options=[
                            {"label": "Mean", "value": "mean"},
                            {"label": "P10 (high)", "value": "p10"},
                            {"label": "P50 (median)", "value": "p50"},
                            {"label": "P90 (low)", "value": "p90"},
                            {"label": "Maximum", "value": "max"},
                            {"label": "Minimum", "value": "min"},
                        ],
                    )
                ],
            ),
            html.Div(
                id={"id": options_uuid, "element": "single_real_options"},
                children=[
                    wcc.Dropdown(
                        label="Realization",
                        id={"id": options_uuid, "element": "realization"},
                        options=[],
                        value=None,
                        multi=False,
                    )
                ],
            ),
        ],
    )



from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from webviz_config import WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from webviz_subsurface._datainput.fmu_input import load_csv
from webviz_subsurface._utils.unique_theming import unique_colors

from ._processing import filter_frame


class RftPlotterDataModel:
    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: Optional[List[str]],
        formations: Path = None,
        faultlines: Path = None,
        obsdata: Path = None,
        csvfile_rft: Path = None,
        csvfile_rft_ert: Path = None,
    ):
        self.formations = formations
        self.faultlines = faultlines
        self.obsdata = obsdata
        self.csvfile_rft = csvfile_rft
        self.csvfile_rft_ert = csvfile_rft_ert

        self.simdf = read_csv(self.csvfile_rft) if csvfile_rft is not None else None
        self.formationdf = read_csv(self.formations) if self.formations else None
        self.faultlinesdf = read_csv(self.faultlines) if self.faultlines else None
        self.obsdatadf = read_csv(self.obsdata) if self.obsdata else None
        self.ertdatadf = pd.DataFrame()

        if csvfile_rft_ert and ensembles:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile_rft_ert" or "ensembles"'
            )

        if csvfile_rft_ert is not None:
            self.ertdatadf = read_csv(self.csvfile_rft_ert)

        if ensembles is not None:
            self.ens_paths = {
                ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
                for ens in ensembles
            }

            try:
                self.simdf = load_csv(self.ens_paths, "share/results/tables/rft.csv")
            except (KeyError, OSError):
                self.simdf = None

            try:
                self.ertdatadf = load_csv(
                    self.ens_paths, "share/results/tables/rft_ert.csv"
                )
            except KeyError as exc:
                raise KeyError(
                    "CSV file for ERT RFT observations/simulations "
                    "(share/results/tables/rft_ert.csv) not found!"
                ) from exc

        self.ertdatadf = self.ertdatadf.rename(
            columns={
                "time": "DATE",
                "is_active": "ACTIVE",
                "isactive": "ACTIVE",
                "well": "WELL",
                "zone": "ZONE",
                "pressure": "SIMULATED",
                "true_vertical_depth": "TVD",
                "measured_depth": "MD",
                "observed": "OBSERVED",
                "obs": "OBSERVED",
                "error": "OBSERVED_ERR",
                "utm_x": "EAST",
                "utm_y": "NORTH",
            }
        )
        self.ertdatadf["DIFF"] = (
            self.ertdatadf["SIMULATED"] - self.ertdatadf["OBSERVED"]
        )
        self.ertdatadf["ABSDIFF"] = abs(
            self.ertdatadf["SIMULATED"] - self.ertdatadf["OBSERVED"]
        )
        self.ertdatadf["YEAR"] = pd.to_datetime(self.ertdatadf["DATE"]).dt.year
        self.ertdatadf = self.ertdatadf.sort_values(by="DATE")
        self.ertdatadf["DATE_IDX"] = self.ertdatadf["DATE"].apply(
            lambda x: list(self.ertdatadf["DATE"].unique()).index(x)
        )
        self.date_marks = self.set_date_marks()
        self.ertdatadf = filter_frame(
            self.ertdatadf,
            {
                "ACTIVE": 1,
            },
        )
        self.ertdatadf["STDDEV"] = self.ertdatadf.groupby(
            ["WELL", "DATE", "ZONE", "ENSEMBLE", "TVD"]
        )["SIMULATED"].transform("std")

    @property
    def well_names(self) -> List[str]:
        return sorted(list(self.ertdatadf["WELL"].unique()))

    @property
    def zone_names(self) -> List[str]:
        return sorted(list(self.ertdatadf["ZONE"].unique()))

    @property
    def dates(self) -> List[str]:
        return sorted(list(self.ertdatadf["DATE"].unique()))

    def date_in_well(self, well: str) -> List[str]:
        df = self.ertdatadf.loc[self.ertdatadf["WELL"] == well]
        return [str(d) for d in list(df["DATE"].unique())]

    @property
    def ensembles(self) -> List[str]:
        return list(self.ertdatadf["ENSEMBLE"].unique())

    @property
    def enscolors(self) -> dict:
        return unique_colors(self.ensembles)

    def set_date_marks(self) -> Dict[str, Dict[str, Any]]:
        marks = {}
        idx_steps = np.linspace(
            start=0,
            stop=self.ertdatadf["DATE_IDX"].max(),
            num=min(4, len(self.ertdatadf["DATE_IDX"].unique())),
            dtype=int,
        )
        date_steps = self.ertdatadf.loc[self.ertdatadf["DATE_IDX"].isin(idx_steps)][
            "DATE"
        ].unique()

        for i, date_index in enumerate(idx_steps):
            marks[str(date_index)] = {
                "label": f"{date_steps[i]}",
                "style": {
                    "white-space": "nowrap",
                    "font-weight": "bold",
                },
            }
        return marks

    @property
    def webviz_store(self) -> List[Tuple[Callable, List[Dict[str, Any]]]]:
        functions: List[Tuple[Callable, List[Dict[str, Any]]]] = [
            (
                read_csv,
                [
                    {"csv_file": path}
                    for path in [
                        self.faultlines,
                        self.formations,
                        self.obsdata,
                        self.csvfile_rft,
                        self.csvfile_rft_ert,
                    ]
                    if path is not None
                ],
            )
        ]
        if self.csvfile_rft_ert is None:
            functions.append(
                (
                    load_csv,
                    [
                        {
                            "ensemble_paths": self.ens_paths,
                            "csv_file": "share/results/tables/rft_ert.csv",
                        },
                    ],
                )
            )
            try:
                load_csv(self.ens_paths, "share/results/tables/rft.csv")
                functions.append(
                    (
                        load_csv,
                        [
                            {
                                "ensemble_paths": self.ens_paths,
                                "csv_file": "share/results/tables/rft.csv",
                            },
                        ],
                    )
                )
            except KeyError:
                pass

        return functions


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: str) -> pd.DataFrame:
    return pd.read_csv(csv_file)



from typing import Callable

import webviz_core_components as wcc


def selections_layout(get_uuid: Callable, ensembles: list) -> wcc.Selectors:
    """Layout for the component input options"""
    controls_uuid = get_uuid("controls")
    return wcc.Selectors(
        id=get_uuid("selections_layout"),
        label="Controls",
        children=[
            wcc.Dropdown(
                label="Ensemble",
                id={"id": controls_uuid, "element": "ensemble"},
                options=[{"label": ens, "value": ens} for ens in ensembles],
                clearable=False,
                value=ensembles[0],
            ),
            wcc.RadioItems(
                label="Mean or realization",
                id={"id": controls_uuid, "element": "tree_mode"},
            ),
        ],
    )



from typing import Callable, List

import webviz_core_components as wcc
from dash import html

from ._business_logic import RftPlotterDataModel


def main_layout(get_uuid: Callable, datamodel: RftPlotterDataModel) -> wcc.Tabs:
    return wcc.Tabs(
        children=[
            wcc.Tab(
                label="RFT Map",
                children=[
                    wcc.FlexBox(
                        children=[
                            wcc.Frame(
                                style={"flex": 1, "height": "87vh"},
                                children=[
                                    map_plot_selectors(get_uuid, datamodel),
                                    formation_plot_selectors(get_uuid, datamodel),
                                ],
                            ),
                            wcc.Frame(
                                style={"flex": 3, "height": "87vh"},
                                color="white",
                                highlight=False,
                                id=get_uuid("map"),
                                children=[],
                            ),
                            wcc.Frame(
                                style={"flex": 3, "height": "87vh"},
                                color="white",
                                highlight=False,
                                id=get_uuid("formations-graph-wrapper"),
                                children=[],
                            ),
                        ]
                    )
                ],
            ),
            wcc.Tab(
                label="RFT misfit per real",
                children=[
                    wcc.FlexBox(
                        children=[
                            wcc.Frame(
                                style={"flex": 1, "height": "87vh"},
                                children=filter_layout(
                                    get_uuid, datamodel, "misfitplot"
                                ),
                            ),
                            wcc.Frame(
                                style={"flex": 6, "height": "87vh"},
                                color="white",
                                highlight=False,
                                id=get_uuid("misfit-graph-wrapper"),
                                children=[],
                            ),
                        ]
                    )
                ],
            ),
            wcc.Tab(
                label="RFT crossplot - sim vs obs",
                children=[
                    wcc.FlexBox(
                        children=[
                            wcc.Frame(
                                style={"flex": 1, "height": "87vh"},
                                children=[
                                    filter_layout(get_uuid, datamodel, "crossplot"),
                                    size_color_layout(get_uuid),
                                ],
                            ),
                            wcc.Frame(
                                style={"flex": 6, "height": "87vh"},
                                color="white",
                                highlight=False,
                                children=[
                                    html.Div(id=get_uuid("crossplot-graph-wrapper")),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
            wcc.Tab(
                label="RFT misfit per observation",
                children=[
                    wcc.FlexBox(
                        children=[
                            wcc.Frame(
                                style={"flex": 1, "height": "87vh"},
                                children=filter_layout(
                                    get_uuid, datamodel, "errorplot"
                                ),
                            ),
                            wcc.Frame(
                                color="white",
                                highlight=False,
                                style={"flex": 6, "height": "87vh"},
                                id=get_uuid("errorplot-graph-wrapper"),
                                children=[],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )


def formation_plot_selectors(
    get_uuid: Callable, datamodel: RftPlotterDataModel
) -> List[html.Div]:
    ensembles = datamodel.ensembles
    well_names = datamodel.well_names
    date_in_well = datamodel.date_in_well
    return wcc.Selectors(
        label="Formation plot settings",
        children=[
            wcc.Dropdown(
                label="Ensemble",
                id=get_uuid("ensemble"),
                options=[{"label": ens, "value": ens} for ens in ensembles],
                value=ensembles[0],
                multi=True,
                clearable=False,
            ),
            wcc.Dropdown(
                label="Well",
                id=get_uuid("well"),
                options=[{"label": well, "value": well} for well in well_names],
                value=well_names[0],
                clearable=False,
            ),
            wcc.Dropdown(
                label="Date",
                id=get_uuid("date"),
                options=[
                    {"label": date, "value": date}
                    for date in date_in_well(well_names[0])
                ],
                clearable=False,
                value=date_in_well(well_names[0])[0],
            ),
            wcc.RadioItems(
                label="Plot simulations as",
                id=get_uuid("linetype"),
                options=[
                    {
                        "label": "Realization lines",
                        "value": "realization",
                    },
                    {
                        "label": "Statistical fanchart",
                        "value": "fanchart",
                    },
                ],
                value="realization",
            ),
            wcc.RadioItems(
                label="Depth option",
                id=get_uuid("depth_option"),
                options=[
                    {
                        "label": "TVD",
                        "value": "TVD",
                    },
                    {
                        "label": "MD",
                        "value": "MD",
                    },
                ],
                value="TVD",
            ),
        ],
    )


def map_plot_selectors(
    get_uuid: Callable, datamodel: RftPlotterDataModel
) -> List[html.Div]:
    ensembles = datamodel.ensembles
    return wcc.Selectors(
        label="Map plot settings",
        children=[
            wcc.Dropdown(
                label="Ensemble",
                id=get_uuid("map_ensemble"),
                options=[{"label": ens, "value": ens} for ens in ensembles],
                value=ensembles[0],
                clearable=False,
            ),
            wcc.Dropdown(
                label="Size points by",
                id=get_uuid("map_size"),
                options=[
                    {
                        "label": "Standard Deviation",
                        "value": "STDDEV",
                    },
                    {
                        "label": "Misfit",
                        "value": "ABSDIFF",
                    },
                ],
                value="ABSDIFF",
                clearable=False,
            ),
            wcc.Dropdown(
                label="Color points by",
                id=get_uuid("map_color"),
                options=[
                    {
                        "label": "Misfit",
                        "value": "ABSDIFF",
                    },
                    {
                        "label": "Standard Deviation",
                        "value": "STDDEV",
                    },
                    {
                        "label": "Year",
                        "value": "YEAR",
                    },
                ],
                value="STDDEV",
                clearable=False,
            ),
            wcc.RangeSlider(
                label="Filter date range",
                id=get_uuid("map_date"),
                min=datamodel.ertdatadf["DATE_IDX"].min(),
                max=datamodel.ertdatadf["DATE_IDX"].max(),
                value=[
                    datamodel.ertdatadf["DATE_IDX"].min(),
                    datamodel.ertdatadf["DATE_IDX"].max(),
                ],
                marks=datamodel.date_marks,
            ),
        ],
    )


def filter_layout(
    get_uuid: Callable, datamodel: RftPlotterDataModel, tab: str
) -> List[wcc.Selectors]:
    """Layout for shared filters"""
    ensembles = datamodel.ensembles
    well_names = datamodel.well_names
    zone_names = datamodel.zone_names
    dates = datamodel.dates
    return wcc.Selectors(
        label="Selectors",
        children=[
            wcc.SelectWithLabel(
                label="Ensembles",
                size=min(4, len(ensembles)),
                id=get_uuid(f"ensemble-{tab}"),
                options=[{"label": name, "value": name} for name in ensembles],
                value=ensembles,
                multi=True,
            ),
            wcc.SelectWithLabel(
                label="Wells",
                size=min(20, len(well_names)),
                id=get_uuid(f"well-{tab}"),
                options=[{"label": name, "value": name} for name in well_names],
                value=well_names,
                multi=True,
            ),
            wcc.SelectWithLabel(
                label="Zones",
                size=min(10, len(zone_names)),
                id=get_uuid(f"zone-{tab}"),
                options=[{"label": name, "value": name} for name in zone_names],
                value=zone_names,
                multi=True,
            ),
            wcc.SelectWithLabel(
                label="Dates",
                size=min(10, len(dates)),
                id=get_uuid(f"date-{tab}"),
                options=[{"label": name, "value": name} for name in dates],
                value=dates,
                multi=True,
            ),
        ],
    )


def size_color_layout(get_uuid: Callable) -> List[html.Div]:
    return wcc.Selectors(
        label="Plot settings",
        children=[
            wcc.Dropdown(
                label="Color by",
                id=get_uuid("crossplot_color"),
                options=[
                    {
                        "label": "Misfit",
                        "value": "ABSDIFF",
                    },
                    {
                        "label": "Standard Deviation",
                        "value": "STDDEV",
                    },
                ],
                value="STDDEV",
                clearable=False,
            ),
            wcc.Dropdown(
                label="Size by",
                id=get_uuid("crossplot_size"),
                options=[
                    {
                        "label": "Standard Deviation",
                        "value": "STDDEV",
                    },
                    {
                        "label": "Misfit",
                        "value": "ABSDIFF",
                    },
                ],
                value="ABSDIFF",
                clearable=False,
            ),
        ],
    )



from typing import Callable

import webviz_core_components as wcc
from dash import html

from .filters_view import filters_layout
from .options_view import options_layout
from .selections_view import selections_layout


def main_view(get_uuid: Callable, ensembles: list) -> wcc.FlexBox:
    """Main layout"""
    return wcc.FlexBox(
        id=get_uuid("layout"),
        children=[
            wcc.FlexColumn(
                flex=1,
                children=[
                    wcc.Frame(
                        style={"height": "82vh"},
                        children=[
                            selections_layout(get_uuid, ensembles),
                            options_layout(get_uuid),
                            filters_layout(get_uuid),
                        ],
                    )
                ],
            ),
            wcc.FlexColumn(
                flex=5,
                children=[
                    wcc.Frame(
                        style={"height": "82vh"},
                        highlight=False,
                        color="white",
                        children=html.Div(
                            id=get_uuid("grouptree_wrapper"),
                        ),
                    )
                ],
            ),
        ],
    )



from typing import Callable

import webviz_core_components as wcc


def filters_layout(get_uuid: Callable) -> wcc.Selectors:
    """The filters part of the menu"""
    filters_uuid = get_uuid("filters")
    return wcc.Selectors(
        id=get_uuid("filters_layout"),
        label="Filters",
        children=[
            wcc.SelectWithLabel(
                label="Prod/Inj/Other",
                id={"id": filters_uuid, "element": "prod_inj_other"},
                options=[
                    {"label": "Production", "value": "prod"},
                    {"label": "Injection", "value": "inj"},
                    {"label": "Other", "value": "other"},
                ],
                value=["prod", "inj", "other"],
                multi=True,
                size=3,
            )
        ],
    )



import datetime

import pandas as pd
import pytest

from webviz_subsurface.plugins._group_tree.group_tree_data import add_nodetype_for_ens

ADD_NODETYPE_CASES = [
    # Group leaf nodes:
    # NODE1 has summary data>0 and will be classified as prod and inj
    # NODE2 has summary data==0 and will be classified as other
    # NODE3 has no summary data and will be classified as other
    # FIELD and TMPL are classified as all three types
    pytest.param(
        pd.DataFrame(
            columns=["DATE", "CHILD", "KEYWORD", "PARENT"],
            data=[
                ["2000-01-01", "FIELD", "GRUPTREE", None],
                ["2000-01-01", "TMPL", "GRUPTREE", "FIELD"],
                ["2000-01-01", "NODE1", "GRUPTREE", "TMPL"],
                ["2000-01-01", "NODE2", "GRUPTREE", "TMPL"],
                ["2000-01-01", "NODE3", "GRUPTREE", "TMPL"],
            ],
        ),
        pd.DataFrame(
            columns=["DATE", "GGPR:NODE1", "GGIR:NODE1", "GGPR:NODE2", "GGIR:NODE2"],
            data=[
                [datetime.date(2000, 1, 1), 1, 1, 0, 0],
            ],
        ),
        pd.DataFrame(
            columns=[
                "DATE",
                "CHILD",
                "KEYWORD",
                "PARENT",
                "IS_PROD",
                "IS_INJ",
                "IS_OTHER",
            ],
            data=[
                ["2000-01-01", "FIELD", "GRUPTREE", None, True, True, True],
                ["2000-01-01", "TMPL", "GRUPTREE", "FIELD", True, True, True],
                ["2000-01-01", "NODE1", "GRUPTREE", "TMPL", True, True, False],
                ["2000-01-01", "NODE2", "GRUPTREE", "TMPL", False, False, True],
                ["2000-01-01", "NODE3", "GRUPTREE", "TMPL", False, False, True],
            ],
        ),
        id="add-nodetype-for-group-leaf-nodes",
    ),
    # Well leaf nodes:
    # WELL1 has WSTAT==1 and will be classified as producer
    # WELL2 has WSTAT==2 and will be classified as injector
    # WELL3 has first WSTAT==1 and then WSTAT==2 and will be classified as both
    # WELL4 has WSTAT==0 and will be classified as other
    # TMPL_A is classified as prod and inj
    # TMPL_B is classified as other
    pytest.param(
        pd.DataFrame(
            columns=["DATE", "CHILD", "KEYWORD", "PARENT"],
            data=[
                ["2000-01-01", "FIELD", "GRUPTREE", None],
                ["2000-01-01", "TMPL_A", "GRUPTREE", "FIELD"],
                ["2000-01-01", "TMPL_B", "GRUPTREE", "FIELD"],
                ["2000-01-01", "WELL1", "WELSPECS", "TMPL_A"],
                ["2000-01-01", "WELL2", "WELSPECS", "TMPL_A"],
                ["2000-01-01", "WELL3", "WELSPECS", "TMPL_A"],
                ["2000-01-01", "WELL4", "WELSPECS", "TMPL_B"],
            ],
        ),
        pd.DataFrame(
            columns=[
                "DATE",
                "WSTAT:WELL1",
                "WSTAT:WELL2",
                "WSTAT:WELL3",
                "WSTAT:WELL4",
            ],
            data=[
                [datetime.date(2000, 1, 1), 1, 2, 1, 0],
                [datetime.date(2000, 2, 1), 1, 2, 2, 0],
            ],
        ),
        pd.DataFrame(
            columns=[
                "DATE",
                "CHILD",
                "KEYWORD",
                "PARENT",
                "IS_PROD",
                "IS_INJ",
                "IS_OTHER",
            ],
            data=[
                ["2000-01-01", "FIELD", "GRUPTREE", None, True, True, True],
                ["2000-01-01", "TMPL_A", "GRUPTREE", "FIELD", True, True, False],
                ["2000-01-01", "TMPL_B", "GRUPTREE", "FIELD", False, False, True],
                ["2000-01-01", "WELL1", "WELSPECS", "TMPL_A", True, False, False],
                ["2000-01-01", "WELL2", "WELSPECS", "TMPL_A", False, True, False],
                ["2000-01-01", "WELL3", "WELSPECS", "TMPL_A", True, True, False],
                ["2000-01-01", "WELL4", "WELSPECS", "TMPL_B", False, False, True],
            ],
        ),
        id="add-nodetype-for-well-leaf-nodes",
    ),
]


@pytest.mark.parametrize("gruptree, smry, expected", ADD_NODETYPE_CASES)
def test_add_nodetype(gruptree, smry, expected):
    """Test functionality of the add_nodetype_for_ens function"""
    columns_to_check = [
        "DATE",
        "CHILD",
        "KEYWORD",
        "PARENT",
        "IS_PROD",
        "IS_INJ",
        "IS_OTHER",
    ]
    output = add_nodetype_for_ens(gruptree, smry)

    pd.testing.assert_frame_equal(output[columns_to_check], expected[columns_to_check])
